
#include <arch/arch.h>
#include <drivers/serial/serial.h>

#define NIBBLES_PER_WORD	(8)

#define LSR_FE			0x08	// Framing Error
#define LSR_PE			0x04	// Parity Error
#define LSR_OE			0x02	// Overrun Error



void serial_init(void)
{
	// GP39, GP40, GP41 UART(10) .   Ѵ
	GPIO_REG(GAFR1_L) |= 0x000A8000;
	GPIO_REG(GPDR1) |= 0x00000380;

	// 8-bit, 1 stop, no parity . 
	FFUART_REG(FFLCR) = 0x00000003;

	// Reset tx, rx FIFO. clear. FIFO enable
	FFUART_REG(FFFCR) = 0x00000007;

	// UART Enable Interrupt
	FFUART_REG(FFIER) = 0x00000040;

	// DLAB set=latch registers, DLAB clear= . Ϲ Ʈ
	FFUART_REG(FFLCR) |= 0x00000080;

	// baud rate . 
	FFUART_REG(FFDLL) = 0x00000008;

	// DLAB clear, . Ϲ Ʈ ȯ
	FFUART_REG(FFLCR) &= 0xFFFFFF7F;

	// Transmit Shift Register, Transmit Holding Register, FIFO 
	// . Ÿ  ٸ
	while ((FFUART_REG(FFLSR) & 0x00000040) == 0);

	return;
}

/*
 * check if there is a character available to read. returns 1 if there
 * is a character available, 0 if not, and negative error number on
 * failure */
int serial_poll(void)
{
	if (FFUART_REG(FFLSR) & (LSR_FE | LSR_PE | LSR_OE))
		return -1;

	if ((FFUART_REG(FFLSR) & 0x00000001) == 0)
		return 0;
	else
		return 1;
}

/* 
 * flush serial input queue. returns 0 on success or negative error
 * number otherwise
 */
int serial_flush_input(void)
{
	volatile int32u tmp;

	while (FFUART_REG(FFLSR) & 0x00000001) {
		if (FFUART_REG(FFLSR) & (LSR_FE | LSR_PE | LSR_OE))
			return -1;
		tmp = FFRBR;
	}

	return 0;
}

/*
 * flush output queue. returns 0 on success or negative error number
 * otherwise
 */
int serial_flush_output(void)
{
	while ((FFUART_REG(FFLSR) & 0x00000040) == 0);

	return 0;
}

/*
 * read one character from the serial port. return character (between
 * 0 and 255) on success, or negative error number on failure. this
 * function is blocking */
int serial_read(void)
{
	int rv;

	for (;;) {
		rv = serial_poll();

		if (rv > 0)	return (FFUART_REG(FFRBR) & 0xFF);
		if (rv < 0)	return (rv);
	}
}

/* write character to serial port and replace all \n characters by a
 * \n\r sequence. return 0 on success, or negative error number on
 * failure. this function is blocking
 */
int serial_write_raw(const char c)
{
	while ((FFUART_REG(FFLSR) & 0x00000020) == 0);

	FFUART_REG(FFTHR) = ((int32u)c & 0xFF);

	return 0;
}

int serial_write(const char c)
{
	while ((FFUART_REG(FFLSR) & 0x00000020) == 0);

	FFUART_REG(FFTHR) = ((int32u)c & 0xFF);

	if (c == '\n')
		serial_write_raw('\r');

	return 0;
}









/*
 * Write a null terminated string to the serial port.
 */
void SerialOutputString(const char *s)
{
	while (*s != 0)
		serial_write(*s++);
} 

/*
 * Write the argument of the function in hexadecimal to the serial
 * port. If you want "0x" in front of it, you'll have to add it
 * yourself.
 */
void SerialOutputHex(const int32u h)
{
	char c;
	int i;
	
	for (i = NIBBLES_PER_WORD - 1; i >= 0; i--) {
		c = (char)((h >> (i * 4)) & 0x0f);

		if (c > 9)
			c += ('a' - 10);
		else
			c += '0';

		serial_write(c);
	}
}

/*
 * Write the argument of the function in decimal to the serial port.
 * We just assume that each argument is positive (i.e. unsigned).
 */
void SerialOutputDec(const int32u d)
{
	int leading_zero = 1;
	int32u divisor, result, remainder;

	remainder = d;

	for (divisor = 1000000000; divisor > 0; divisor /= 10) {
		result = remainder / divisor;
		remainder %= divisor;

		if (result != 0 || divisor == 1)
			leading_zero = 0;

		if (leading_zero == 0)
			serial_write((char)(result) + '0');
	}
}

/*
 * Write a block of data to the serial port. Similar to
 * SerialOutputString(), but this function just writes the number of
 * characters indicated by bufsize and doesn't look at termination
 * characters.
 */
void SerialOutputBlock(const char *buf, int bufsize)
{
	while (bufsize--)
		serial_write(*buf++);
}


/*
 * read a string with maximum length len from the serial port
 * using a timeout of timeout seconds
 *
 * len is the length of array s _including_ the trailing zero,
 * the function returns the number of bytes read _excluding_
 * the trailing zero
 */
int  SerialInputString(char *s, const int len)
{
	int c;
	int i;

	for (i = 0; i < len - 1; i++) {
		/* try to get a byte from the serial port */
		while (serial_poll() == 0)
			;

		c = serial_read();

		/* check for errors */
		if(c < 0) {
			s[i] = '\0';
			return c;
		}

		if((c == '\r') || (c == '\n')) {
			s[i] = '\0';
			serial_write('\n');
			return (i);
		}
		else if (c == '\b') {
			if (i > 0) {
				i--;
				SerialOutputString("\b \b");
			}
		} else {
			s[i] = (char) c;
			serial_write(c);
		}
	}

	return (i);
}

/* 
 * SerialInputBlock(): almost the same as SerialInputString(), but
 * this one just reads a block of characters without looking at
 * special characters.
 */
int  SerialInputBlock(char *buf, int bufsize)
{
	int c;
	int i;

	for (i = 0; i < bufsize - 1; i++) {
		/* try to get a byte from the serial port */
		while (serial_poll() == 0)
			;

		c = serial_read();

		/* check for errors */
		if(c < 0) {
			return c;
		}

		buf[i] = (char) c;
	}

	return (i);
}

