#ifndef _BIT_OP_H
#define _BIT_OP_H

#include <arch/types.h>

#define	BITS_PER_LONG		(32)
#define BITOP_MASK(nr)		(1UL << ((nr) % BITS_PER_LONG))
#define BITOP_WORD(nr)		((nr) / BITS_PER_LONG)

/**
 * SetBit - Set a bit in memory
 * @nr: the bit to set
 * @addr: the address to start counting from
 *
 * Unlike set_bit(), this function is non-atomic and may be reordered.
 * If it's called on the same region of memory simultaneously, the effect
 * may be that only one operation succeeds.
 */
static inline void SetBit(int nr, volatile int32u *addr)
{
	int32u mask = BITOP_MASK(nr);
	int32u *p = ((int32u *)addr) + BITOP_WORD(nr);

	*p  |= mask;
}

static inline void ClearBit(int nr, volatile int32u *addr)
{
	int32u mask = BITOP_MASK(nr);
	int32u *p = ((int32u *)addr) + BITOP_WORD(nr);

	*p &= ~mask;
}

/**
 * ChangeBit - Toggle a bit in memory
 * @nr: the bit to change
 * @addr: the address to start counting from
 *
 * Unlike ChangeBit(), this function is non-atomic and may be reordered.
 * If it's called on the same region of memory simultaneously, the effect
 * may be that only one operation succeeds.
 */
static inline void ChangeBit(int nr, volatile int32u *addr)
{
	int32u mask = BITOP_MASK(nr);
	int32u *p = ((int32u *)addr) + BITOP_WORD(nr);

	*p ^= mask;
}

/**
 * TestAndSet - Set a bit and return its old value
 * @nr: Bit to set
 * @addr: Address to count from
 *
 * This operation is non-atomic and can be reordered.
 * If two examples of this operation race, one can appear to succeed
 * but actually fail.  You must protect multiple accesses with a lock.
 */
static inline int32s TestAndSet(int nr, volatile int32u *addr)
{
	int32u mask = BITOP_MASK(nr);
	int32u *p = ((int32u *)addr) + BITOP_WORD(nr);
	int32u old = *p;

	*p = old | mask;
	return (old & mask) != 0;
}

/**
 * TestAndClear - Clear a bit and return its old value
 * @nr: Bit to clear
 * @addr: Address to count from
 *
 * This operation is non-atomic and can be reordered.
 * If two examples of this operation race, one can appear to succeed
 * but actually fail.  You must protect multiple accesses with a lock.
 */
static inline int32s TestAndClear(int nr, volatile int32u *addr)
{
	int32u mask = BITOP_MASK(nr);
	int32u *p = ((int32u *)addr) + BITOP_WORD(nr);
	int32u old = *p;

	*p = old & ~mask;
	return (old & mask) != 0;
}

/* WARNING: non atomic and it can be reordered! */
static inline int32s TestAndChange(int nr,
					    volatile int32u *addr)
{
	int32u mask = BITOP_MASK(nr);
	int32u *p = ((int32u *)addr) + BITOP_WORD(nr);
	int32u old = *p;

	*p = old ^ mask;
	return (old & mask) != 0;
}

/**
 * TestBit - Determine whether a bit is set
 * @nr: bit number to test
 * @addr: Address to start counting from
 */
static inline int32s TestBit(int nr, const volatile int32u *addr)
{
	return 1UL & (addr[BITOP_WORD(nr)] >> (nr & (BITS_PER_LONG-1)));
}

#endif /* _BIT_OP_H*/
